from ..utils import register

from aiohttp import ClientError, ClientSession, TCPConnector
from html.parser import HTMLParser, unescape

import asyncio
import re

default_charset = 'utf-8'

max_loop = 5
timeout = 2
max_bytes = 10240
re_charsets = (
    re.compile(rb"< *meta.+?charset=([\"']?)([\w\s\_-]+)\1[^>]*>"),
    re.compile(
        rb"< *meta.+?content=([\"']?)[^;]*; +charset=([\w\s\_-]+)\1[^>]*>"),
)
re_space = re.compile(r'\s+')


class TitleParser(HTMLParser):
    title_tag = False
    first = False
    title = None
    charset = "utf-8"

    def handle_starttag(self, tag, attrs):
        if self.first:
            return
        if tag == "title":
            self.title_tag = True
            self.first = True

    def handle_data(self, data):
        if self.title_tag:
            self.title = (self.title or '') + data

    def handle_endtag(self, tag):
        if tag == "title":
            self.title_tag = False

    @classmethod
    def get_title(cls, data):
        parser = cls(convert_charrefs=True)

        try:
            parser.feed(data)
        except:  #many bugs lol
            return None

        title = parser.title

        if title is None:
            return None

        title = title.strip()

        if title:
            return re_space.sub(' ', unescape(title))


@asyncio.coroutine
def get_response(address, headers=None, redirect=True,
                 ssl=False, session=None):

    if session is None:
        session = ClientSession(
            connector=TCPConnector(
                verify_ssl=ssl,
                conn_timeout=timeout
            )
        )
        session.max_redirects = max_loop

    try:
        resp = yield from session.get(
            address,
            #headers=headers,
            allow_redirects=redirect
        )
    except ClientError:
        return None, None

    if resp.status != 200:
        return None, None

    raw = yield from resp.content.read(max_bytes)
    charset = default_charset
    for re_charset in re_charsets:
        charset_match = re_charset.search(raw)
        if charset_match:
            charset = charset_match.group(2).decode()
            break

    try:
        msg = raw.decode(charset, errors='replace')
    except LookupError:
        return None, None

    resp.close()
    return msg, resp


@asyncio.coroutine
def check_title(bot, address, ssl):
    msg, resp = yield from get_response(address, ssl)

    if resp is None:
        return

    headers = resp.headers
    if "content-type" in headers:
        cont_type = headers["content-type"]
        if cont_type.startswith("text/html") and msg:
            title = TitleParser.get_title(msg)
            if title:
                bot.say("⚡ %s" % title)


@register(reg=r'http(s?)://(\S+)|(www\.\S+)')
def title(bot, ssl: lambda s: s == 's', address, address_another):
    address = address or address_another

    if not address.startswith('http'):
        address = "http://" + address

    asyncio.async(check_title(bot, address, ssl))



'''
Useful decorators and function
'''
import re
import inspect


def create_help(cmd, txt):
    pass


def init_plugin(func):
    func.is_initial = True
    return func

class register(object):
    default_kwargs = {
        "on_private": True,
        "on_channel": True,
        "event": "msg",
        "is_reg": False,
        "reg": None,
        "cmd": None,
        "next": False,
        "name": None,
        "priority": 0,
        "block": True,
        "first_char": '.',
        "admin_required": False,
    }

    def __init__(self, **kwargs):
        """
        :param cmd: Command name. Attributes must have annotations!
        :param reg: regexp without command name.
        :param on_private
        :param on_channel
        :param name alternative name of event. If is not defined then name is function name
        :param next: if is true, accepting another matching commands
        :param priority: 0 is a smallest priority
        :param first_char: char before command name
        :param admin_required: if is true - check permissions
        """

        self.kwargs = self.default_kwargs.copy()

        event_type = kwargs.get("event", "msg")
        reg = kwargs.get("reg", None)
        if reg:
            kwargs['compiled_reg'] = re.compile(reg)
            kwargs["is_reg"] = True
        if event_type == "msg":
            if not(reg or kwargs.get("cmd", None)):
                raise ValueError("Must be 'cmd' or 'reg' argument")
        else:
            raise ValueError("event argument is invalid")

        self.kwargs.update(kwargs)

    def __call__(self, func):
        func.__dict__.update(self.kwargs)
        func.is_bot_event = True
        module_name = func.__module__.split('.')[-1]

        func.name = '%s.%s' % (module_name, func.name or func.__name__.lower())

        if func.event == "msg":
            argspec = inspect.getfullargspec(func)
            if argspec.args:  # count require argument
                len_defaults = len(argspec.defaults) if argspec.defaults else 0
                func.max_args = len(argspec.args) - len_defaults - 1
            else:
                func.max_args = -1

            #events["msg"][func.name] = func
            #modules[module_name].append(func)

        return func



#!/usr/bin/python3

#from plugin import Plugin
from .. import config
import random
from ..utils import register

"""
@register(reg=r'^' + config.nick + r'!$')
def ping(bot):
    bot.say(bot.user.nick + '!')


@register(reg=r'(?i)^' + config.nick + r': +(?:(?:Cze[sś][cć])|(?:Siema(?:sz)?)|(?:He[jy])|(?:!+))[!.]? *$')
def ping2(bot):
    bot.reply(random.choice((
                'Hey', 'Siema', '?', 'czeeeść'
                                )))

"""
@register(cmd='utf8')
def utf(bot):
    """show utf-8 chars"""
    bot.reply("żółw,lwiątko,kurczę")



from ..utils import register
from ..utils.types import regexp

from random import randint


def min_max(num):
    return min(100, max(-100, int(num)))


@register(cmd='roll')
def roll(bot, value:regexp(r'^(\d+)(?:d(-?\d+)(?::(-?\d+))?)?$')):
    """
    usage: .roll 3 || .roll 2d7 || .roll
    """
    many, a, b = value.groups()
    many = int(many)
    if many > 50:
        bot.say('too many steps (%s)!' % many)
        return

    if a is None:
        a, b = 1, 9
    elif b is None:
        a, b = 1, min_max(a)
    else:
        a, b = min_max(a), min_max(b)
        a, b = min(a, b), max(a, b)

    results = [randint(a, b) for i in range(many)]

    bot.reply(
        '({results}) = {sum}'.format(
            results='+'.join(str(x) for x in results),
            sum=sum(results)
        )
    )



#!/usr/bin/python3

import asyncio
#import pyejdb
import re
import datetime

from .message_controller import MessageController
from .sender import IrcSender

re_message = re.compile(
    r'(?:(:[^ ]+) )?'
    r'([^:\r]+)'
    r'(?: :([^\r]+))?'
    r'\r\n'
)


class IrcClient(asyncio.Protocol, IrcSender):

    ready = False
    config = None
    importer = None
    db = None
    connection_lost_future = None
    ping_counter = 0

    def __init__(self, config, connection_lost_future):
        IrcSender.__init__(self)
        asyncio.Protocol.__init__(self)
        self.config = config
        #self.db = pyejdb.EJDB(
        #    self.config.get('main', 'database_path'),
        #    pyejdb.DEFAULT_OPEN_MODE | pyejdb.JBOTRUNC
        #)
        self.importer = self.config.getmodule('main', 'importer')(self)
        self.importer.load_all()
        self.connection_lost_future = connection_lost_future

    def connection_made(self, transport):
        config = self.config['main']
        self.transport = transport
        self.send('PASS', config['password'])
        self.send('NICK', config['nick'])
        self.send_msg('USER', config['ircname'], 'wr', '*', config['realname'])

    def data_received(self, raw_messages):
        codecs = self.config.getlist('main', 'codecs')
        for message in self._parse_raw_messages(raw_messages, codecs):
            MessageController(self, message).execute_message()

    def connection_lost(self, exc):
        self.connection_lost_future.set_exception(exc)
        self.importer.cancel_tasks()

    @staticmethod
    def _parse_raw_messages(raw_messages, codecs=('utf-8',)):
        for codec in codecs:
            try:
                encoded_data = raw_messages.decode(codec)
            except UnicodeDecodeError:
                continue
            else:
                break
        else:
            return

        for match in re_message.finditer(encoded_data):
            prefix, messages, long_message = match.groups()
            data = [prefix] if prefix else []
            data += messages.split()
            if long_message:
                data.append(long_message)
            yield data


def connect(config):
    loop = asyncio.get_event_loop()
    try:
        while True:
            main_config = config['main']
            future = asyncio.Future()
            factory = lambda: IrcClient(config, future)
            coro = loop.create_connection(
                factory, main_config['host'], main_config['port']
            )
            loop.run_until_complete(coro)
            loop.run_until_complete(future)
    finally:
        loop.close()



from ..utils.event_loop import loop
from ..format import color, bold

import requests
import asyncio

API_URL = "https://api.github.com/repos/{user}/{repo}/{action}"
state_colors = {
    'open': color.green,
    'closed': color.red,
    'reopened': color.green,
    'subscribed': color.green,
    'merged': color.green,
    'assigned': color.purple,
    'unassigned': color.purple,
    'milestoned': color.olive,
    'demilestoned': color.olive,
    'locked': color.red,
    'unlocked': color.green,
    'renamed': color.dark_gray
}


class NothingChangeException(Exception):
    pass


@loop(time_config_key="commits_time", default=60)
def commits(irc_protocol, plugin, config):
    try:
        data = github_action('commits', plugin, config)
    except (requests.HTTPError, NothingChangeException):
        return

    msg = "‼ {commit} '{msg}' by {committer} {url}".format(
        commit=bold('COMMIT'),
        msg=strip(data['commit']['message']),
        committer=data['commit']['author']['name'],
        url=data['html_url']
    )

    for chan in config['channels'].split(','):
        irc_protocol.say(chan.strip(), msg)


@loop(time_config_key="events_time", default=60)
def events(irc_protocol, plugin, config):
    try:
        data = github_action('issues/events', plugin, config)
    except (requests.HTTPError, IndexError, NothingChangeException):
        return

    issue = data['issue']
    state = data['event']
    state_color = state_colors.get(state, color.black)
    assignee = issue['assignee']
    assignee_login = assignee['login'] if assignee is not None else None
    msg = ("‼ {issue} {labels} '{title}' by {user} "
           "{assignee} {state} {url}").format(
        issue=bold('ISSUE'),
        labels=' '.join('[%s]' % label['name'] for label in issue['labels']),
        title=strip(issue['title']),
        state=bold(color(' %s ' % state, color.white, state_color)),
        user=issue['user']['login'],
        assignee="assigned to %s" % assignee_login if assignee_login else '',
        url=issue['html_url'],
    )

    for chan in config['channels'].split(','):
        irc_protocol.say(chan.strip(), msg)


@loop(time_config_key="comments_time", default=60)
def comments(irc_protocol, plugin, config):
    try:
        data = github_action('issues/comments', plugin, config)
    except (requests.HTTPError, IndexError, NothingChangeException):
        return

    msg = "‼ {issue} by {nick} '{msg}' {url}".format(
        issue=bold('COMMENT'),
        msg=strip(data['body']),
        nick=data['user']['login'],
        url=data['html_url'],
    )

    for chan in config['channels'].split(','):
        irc_protocol.say(chan.strip(), msg)


def strip(msg):
    lines = msg.splitlines()
    line = lines[0]
    if len(lines) == 0:
        return ''
    elif len(lines) == 1:
        return line if len(line) < 50 else line[:50] + "…"
    else:
        return line[:50] + "…"


def github_action(action, plugin, config, sort_by='updated'):
    url = API_URL.format(
        api=API_URL,
        user=config['user'],
        repo=config['repository'],
        action=action,
    )
    query = {'sort': sort_by, 'direction': 'desc'}
    headers = {'Accept': 'application/vnd.github.v3+json'}
    etag = plugin.temp.get(action)
    if etag is not None:
        headers['If-None-Match'] = etag

    req = requests.get(url, params=query, headers=headers, timeout=5)
    req.raise_for_status()
    plugin.temp[action] = req.headers.get('etag')
    if req.status_code == 304 or etag is None:  # etag must be cached first
        raise NothingChangeException(etag)
    return req.json()[0]



from . import format

from asyncio import async, coroutine, gather
from asyncio.futures import Future

from .request import RequestBot
from importlib import reload
from collections import defaultdict
from inspect import getcallargs
from datetime import datetime, timedelta

import re
import traceback
import sys

#parse args
re_split = re.compile(r' *(?:(\w+)= *)?(?:"([^"]+)"|(\S+))')


class Plugin(list):

    __slots__ = ('temp', 'name', 'module_path', 'future')

    def __init__(self, name, module_path, funcs):
        super().__init__(funcs)
        self.temp = {}
        self.name = name
        self.module_path = module_path
        self.future = Future()

    def __repr__(self):
        return "Plugin('%s')" % self.name


class ExecutedCounter(object):

    __slots__ = ('last_time', 'counter')

    def __init__(self):
        self.last_time = datetime.now()
        self.counter = 0

    def inc(self):
        self.counter += 1


class ModuleManager(object):

    protocol = None
    plugins = None
    executed_counters = None

    def __init__(self, protocol):
        self.protocol = protocol
        self.plugins = {}
        self.executed_counters = defaultdict(ExecutedCounter)

    @property
    def config(self):
        return self.protocol.config

    def load_all(self):
        for name, module_path in self.config.items('plugins'):
            self.load(name, module_path)

    def cancel_tasks(self):
        for plugin in self.plugins.values():
            plugin.future.cancel()

    def load(self, name, module_path):
        path, module_name = module_path.rsplit(".", 1)
        module = __import__(module_path, globals(), locals(), [module_name])
        reload(module)
        self.plugins[name] = plugin = Plugin(
            name, module_path,
            [
                obj for obj in module.__dict__.values()
                if getattr(obj, 'is_bot_event', False) is True
            ]
        )

        initials_funcs = (
            func for func in module.__dict__.values()
            if getattr(func, 'is_initial', False) is True
        )

        for initial_func in initials_funcs:
            initial_func(plugin, self.protocol)

        future = plugin.future
        loops = [
            func(self.protocol, plugin, future)
            for func in module.__dict__.values()
            if getattr(func, 'is_loop', False) is True
        ]

        for loop in loops:
            async(loop)

    def reload(self, name):
        plugin = self.plugins[name]
        plugin.future.cancel()
        self.load(name, plugin.module_path)

    def remove(self, name):
        plugin = self.plugins.pop(name)
        plugin.future.cancel()

    @coroutine
    def execute(self, channel, user, msg):
        cmd, text = self.get_cmd_and_text(msg)
        private = channel[0] != "#"

        if cmd is not None:
            self.execute_command(cmd, text, private, channel, user)

        self.execute_regs(msg, private, channel, user)

    def execute_command(self, cmd, text, private, channel, user):
        plugin, func = self.find_command(cmd, private, channel)
        if func is None:
            return

        if text:
            args, kwargs = get_args_from_text(text, func.max_args)
        else:
            args = []
            kwargs = {}

        self.execute_func(func, plugin, private, channel, user, args, kwargs)

    def execute_regs(self, msg, private, channel, user):
        for plugin, func, matches in self.find_regs(msg, private):
            for match, _ in zip(matches, range(3)):
                args = match.groups()
                kwargs = match.groupdict()
                self.execute_func(func, plugin, private, channel, user,
                                  args, kwargs)

    def find_function_in_plugin_with_name(self, private=None):
        return (
            (plugin, func)
            for name, plugin in self.plugins.items()
            for func in plugin
            if private is None
            or (not private and func.on_channel)
            or (private and func.on_private)
        )

    def find_command(self, cmd, private=None, channel=None):

        def is_good(plugin, func):
            cfg = self.get_plugin_cfg(plugin.name)

            if func.cmd.format(**cfg) != cmd:
                return False

            if private or channel is None:
                return True

            def get_list(key):
                return [
                    x for x in (x.strip() for x in cfg.get(key, '').split(','))
                    if x
                ]

            whitelist = get_list('whitelist')
            if whitelist:
                return channel in whitelist

            blacklist = get_list('blacklist')
            return channel not in blacklist

        return next(
            (
                (plugin, func) for plugin, func
                in self.get_commands(private)
                if is_good(plugin, func)
            ),
            (None, None)
        )

    def get_commands(self, private=None):
        return (
            (plugin, func) for plugin, func
            in self.find_function_in_plugin_with_name(private)
            if func.is_reg is False
        )

    def find_regs(self, msg, private):
        return (
            (plugin, func, func.compiled_reg.finditer(msg))
            for plugin, func in self.find_function_in_plugin_with_name(private)
            if func.is_reg is True
        )

    def get_plugin_cfg(self, name):
        cfg = self.config
        plugin_name = 'plugin:%s' % name
        conf = {}
        if cfg.has_section(plugin_name):
            conf.update(cfg.items(plugin_name))

        conf.update(
            __nick__=cfg.get('main', 'nick')
        )

        return conf

    def get_cmd_and_text(self, msg):
        cfg = self.config['main']
        startswith = lambda key: msg.startswith(cfg[key])
        if startswith('command-prefix'):
            data = msg.split(maxsplit=1)
            cmd = data[0][1:]
        elif startswith('nick'):
            data = msg.split(maxsplit=2)[1:]
            cmd = data[0]
        else:
            return None, msg
        text = data[1] if len(data) > 1 else ""
        return cmd, text

    def execute_func(self, func, plugin, private, channel, user, args, kwargs):

        if func.block and self.is_blocked(user):
            return  # filtr antyspamowy, nienawidze was

        bot = RequestBot(
            protocol=self.protocol,
            private=private,
            chan=channel if not private else None,
            config=self.get_plugin_cfg(plugin.name),
            plugin=plugin,
            user=user,
            temp=plugin.temp
        )

        try:
            dict_arg = check_type(args, kwargs, func)
        except:
            return

        async(self._execute_func(func, bot, dict_arg, user, channel))

    @coroutine
    def _execute_func(self, func, bot, args, user, channel):
        if func.admin_required:
            is_admin = yield from bot.is_admin()
            if not is_admin:
                return
        try:
            func(bot, **args)
        except:
            traceback.print_exc(file=sys.stdout)
            tb = traceback.format_exc().split('\n')[-4:-1]
            self.protocol.reply(
                user.nick,
                '{} {} {}'.format(
                    format.bold('ERR:'),
                    tb[2],
                    format.color(tb[0], format.color.red),
                ),
                channel
            )

    def is_blocked(self, user):
        now = datetime.now()
        delta = timedelta(
            seconds=self.config.getint('main', 'time_to_block')
        )

        self.executed_counters = defaultdict(ExecutedCounter, { # cleanups
            key: counter for key, counter in self.executed_counters.items()
            if now - counter.last_time < delta
        })

        nick = user.prefix
        counter = self.executed_counters[nick]

        executed_commands_per_time = self.config.getint(
            'main', 'executed_commands_per_time'
        )

        if counter.counter > executed_commands_per_time:
            return True
        else:
            counter.inc()
            return False


def get_args_from_text(text, max_args):
    args = []
    kwargs = {}
    for arg_group in re_split.finditer(text):
        arg_name, arg_quotes, arg = arg_group.groups()
        arg = arg_quotes or arg

        if arg_name:
            kwargs[arg_name] = arg
        else:
            if max_args > -1:
                args.append(arg)

    if len(args) > max_args != -1:
        last_arg = ' '.join(args[max_args:])
        args = args[0:max_args]
        try:
            args[-1] += ' ' + last_arg
        except IndexError:
            args = [last_arg]

    return args, kwargs


def check_type(args, kwargs, event):
    dict_arg = getcallargs(event, None, *args, **kwargs)
    for key, item in dict_arg.items():
        arg_type = event.__annotations__.get(key, str)
        if item != None:
            dict_arg[key] = arg_type(item)
    del dict_arg['bot']
    return dict_arg



import asyncio


class User(object):
    """User class - has nick, realname, and host"""
    nick = ""
    host = ""
    realname = ""

    def __init__(self, prefix):
        try:
            self.nick, name_host = prefix.split('!')
            self.realname, self.host = name_host.split('@')
        except:
            pass

    @property
    def prefix(self):
        return "%s@%s" % (self.realname, self.host)


class WhoisFuture(asyncio.Future):

    data = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = WhoisData()


class WhoisData(object):

    __slots__ = ('idle', 'account', 'ircname', 'host',
                 'channels', 'realname', 'server', 'nick')

    def __init__(self):
        self.channels = []
        self.idle = 0
        self.nick = None
        self.account = None
        self.ircname = None
        self.host = None
        self.realname = None
